@@ -3,13 +3,15 @@
    Copyright (c) 2005-2006 Matthew Johnson
 
    This program is free software; you can redistribute it and/or modify it
-   under the terms of either the GNU General Public License Version 2 or the
+   under the terms of either the GNU Lesser General Public License Version 2 or the
    Academic Free Licence Version 2.1.
 
    Full licence texts are included in the COPYING file with this program.
 */
 package org.freedesktop.dbus.bin;
 
+import static org.freedesktop.dbus.Gettext._;
+
 import org.freedesktop.DBus;
 import org.freedesktop.dbus.AbstractConnection;
 import org.freedesktop.dbus.BusAddress;
@@ -32,17 +34,16 @@ import java.io.FileOutputStream;
 import java.io.IOException;
 import java.io.PrintWriter;
 import java.lang.ref.WeakReference;
-import java.lang.reflect.Method;
 import java.lang.reflect.InvocationTargetException;
 import java.net.InetAddress;
 import java.net.ServerSocket;
 import java.net.Socket;
+import java.text.MessageFormat;
 import java.util.HashMap;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
-import java.util.Queue;
 import java.util.Vector;
 
 import cx.ath.matthew.debug.Debug;
@@ -132,7 +133,7 @@ public class DBusDaemon extends Thread
          if (Debug.debug) Debug.print(Debug.DEBUG, "enter");
          synchronized (c) {
             if (null != c.unique) 
-               throw new org.freedesktop.DBus.Error.AccessDenied("Connection has already sent a Hello message");
+               throw new org.freedesktop.DBus.Error.AccessDenied(_("Connection has already sent a Hello message"));
             synchronized (unique_lock) {
                c.unique = ":1."+(++next_unique);
             }
@@ -297,6 +298,7 @@ public class DBusDaemon extends Thread
          if (Debug.debug) Debug.print(Debug.DEBUG, "exit");
          return;
       }
+      @SuppressWarnings("unchecked")
       private void handleMessage(Connstruct c, Message m) throws DBusException
       {
          if (Debug.debug) Debug.print(Debug.DEBUG, "enter");
@@ -304,7 +306,7 @@ public class DBusDaemon extends Thread
          if (!(m instanceof MethodCall)) return;
          Object[] args = m.getParameters();
 
-         Class[] cs = new Class[args.length];
+         Class<? extends Object>[] cs = new Class[args.length];
 
          for (int i = 0; i < cs.length; i++)
             cs[i] = args[i].getClass();
@@ -333,10 +335,10 @@ public class DBusDaemon extends Thread
                send(c, new org.freedesktop.dbus.Error("org.freedesktop.DBus", m, DBEe));
             } catch (Exception e) {
                if (Debug.debug && AbstractConnection.EXCEPTION_DEBUG) Debug.print(Debug.ERR, e);
-               send(c,new org.freedesktop.dbus.Error("org.freedesktop.DBus", c.unique, "org.freedesktop.DBus.Error.GeneralError", m.getSerial(), "s", "An error occurred while calling the "+m.getName()+" Method"));
+               send(c,new org.freedesktop.dbus.Error("org.freedesktop.DBus", c.unique, "org.freedesktop.DBus.Error.GeneralError", m.getSerial(), "s", _("An error occurred while calling ")+m.getName()));
             }
          } catch (NoSuchMethodException NSMe) {
-            send(c,new org.freedesktop.dbus.Error("org.freedesktop.DBus", c.unique, "org.freedesktop.DBus.Error.UnknownMethod", m.getSerial(), "s", "This service does not support the "+m.getName()+" Method"));
+            send(c,new org.freedesktop.dbus.Error("org.freedesktop.DBus", c.unique, "org.freedesktop.DBus.Error.UnknownMethod", m.getSerial(), "s", _("This service does not support ")+m.getName()));
          }
 
          if (Debug.debug) Debug.print(Debug.DEBUG, "exit");
@@ -625,13 +627,13 @@ public class DBusDaemon extends Thread
                         && (!(m instanceof MethodCall) 
                            || !"org.freedesktop.DBus".equals(m.getDestination())
                            || !"Hello".equals(m.getName()))) {
-                     send(c,new Error("org.freedesktop.DBus", null, "org.freedesktop.DBus.Error.AccessDenied", m.getSerial(), "s", "You must send a Hello message"));
+                     send(c,new Error("org.freedesktop.DBus", null, "org.freedesktop.DBus.Error.AccessDenied", m.getSerial(), "s", _("You must send a Hello message")));
                   } else {
                      try {
                         if (null != c.unique) m.setSource(c.unique);
                      } catch (DBusException DBe) {
                         if (Debug.debug && AbstractConnection.EXCEPTION_DEBUG) Debug.print(Debug.ERR, DBe);
-                        send(c,new Error("org.freedesktop.DBus", null, "org.freedesktop.DBus.Error.GeneralError", m.getSerial(), "s", "Sending message failed"));
+                        send(c,new Error("org.freedesktop.DBus", null, "org.freedesktop.DBus.Error.GeneralError", m.getSerial(), "s", _("Sending message failed")));
                      }
 
                      if ("org.freedesktop.DBus".equals(m.getDestination())) {
@@ -648,7 +650,7 @@ public class DBusDaemon extends Thread
                            Connstruct dest = names.get(m.getDestination());
 
                            if (null == dest) {
-                              send(c, new Error("org.freedesktop.DBus", null, "org.freedesktop.DBus.Error.ServiceUnknown", m.getSerial(), "s", "The name `"+m.getDestination()+"'does not exist"));
+                              send(c, new Error("org.freedesktop.DBus", null, "org.freedesktop.DBus.Error.ServiceUnknown", m.getSerial(), "s", MessageFormat.format(_("The name `{0}' does not exist"), new Object[] { m.getDestination() })));
                            } else
                               send(dest, m);
                         }
@@ -816,7 +818,7 @@ public class DBusDaemon extends Thread
       // accept new connections
       while (d._run) {
          UnixSocket s = uss.accept();
-         if ((new Transport.SASL()).auth(Transport.SASL.MODE_SERVER, Transport.SASL.AUTH_EXTERNAL, address.getParameter("guid"), s.getOutputStream(), s.getInputStream())) {
+         if ((new Transport.SASL()).auth(Transport.SASL.MODE_SERVER, Transport.SASL.AUTH_EXTERNAL, address.getParameter("guid"), s.getOutputStream(), s.getInputStream(), s)) {
          //   s.setBlocking(false);
             d.addSock(s);
          } else
@@ -836,7 +838,13 @@ public class DBusDaemon extends Thread
       // accept new connections
       while (d._run) {
          Socket s = ss.accept();
-         if ((new Transport.SASL()).auth(Transport.SASL.MODE_SERVER, Transport.SASL.AUTH_EXTERNAL, address.getParameter("guid"), s.getOutputStream(), s.getInputStream())) {
+         boolean authOK=false;
+         try {
+        	 authOK = (new Transport.SASL()).auth(Transport.SASL.MODE_SERVER, Transport.SASL.AUTH_EXTERNAL, address.getParameter("guid"), s.getOutputStream(), s.getInputStream(), null);
+         } catch (Exception e) {
+        	 if (Debug.debug) Debug. print(Debug.DEBUG, e);
+         }
+         if (authOK) {
             d.addSock(s);
          } else
             s.close();
