@@ -3,7 +3,7 @@
    Copyright (c) 2005-2006 Matthew Johnson
 
    This program is free software; you can redistribute it and/or modify it
-   under the terms of either the GNU General Public License Version 2 or the
+   under the terms of either the GNU Lesser General Public License Version 2 or the
    Academic Free Licence Version 2.1.
 
    Full licence texts are included in the COPYING file with this program.
@@ -26,10 +26,12 @@ public interface Profiler extends DBusInterface
       }
    }
    public void array(int[] v);
+   public void stringarray(String[] v);
    public void map(Map<String,String> m);
    public void list(List<String> l);
    public void bytes(byte[] b);
    public void struct(ProfileStruct ps);
+   public void string(String s);
 }
 
 
