@@ -3,7 +3,7 @@
    Copyright (c) 2005-2006 Matthew Johnson
 
    This program is free software; you can redistribute it and/or modify it
-   under the terms of either the GNU General Public License Version 2 or the
+   under the terms of either the GNU Lesser General Public License Version 2 or the
    Academic Free Licence Version 2.1.
 
    Full licence texts are included in the COPYING file with this program.
@@ -17,8 +17,10 @@ public class ProfilerInstance implements Profiler
 {
    public boolean isRemote() { return false; }
    public void array(int[] v) { return; }
+   public void stringarray(String[] v) { return; }
    public void map(Map<String,String> m) { return; }
    public void list(List<String> l) { return; }
    public void bytes(byte[] b) { return; }
    public void struct(ProfileStruct ps) { return; }
+   public void string(String s) { return; }
 }
