@@ -3,7 +3,7 @@
    Copyright (c) 2005-2006 Matthew Johnson
 
    This program is free software; you can redistribute it and/or modify it
-   under the terms of either the GNU General Public License Version 2 or the
+   under the terms of either the GNU Lesser General Public License Version 2 or the
    Academic Free Licence Version 2.1.
 
    Full licence texts are included in the COPYING file with this program.
@@ -19,11 +19,21 @@ import java.util.Map;
 
 import org.freedesktop.dbus.DirectConnection;
 import org.freedesktop.dbus.Path;
-import org.freedesktop.dbus.Transport;
 import org.freedesktop.dbus.UInt16;
 
 public class test_p2p_server implements TestRemoteInterface
 {
+   public int[][] teststructstruct(TestStruct3 in)
+   {
+      List<List<Integer>> lli = in.b;
+      int[][] out = new int[lli.size()][];
+      for (int j = 0; j < out.length; j++) {
+         out[j] = new int[lli.get(j).size()];
+         for (int k = 0; k < out[j].length; k++)
+            out[j][k] = lli.get(j).get(k);
+      }
+      return out;
+   }
    public String getName()
    {
       System.out.println("getName called");
